BODY,H1,H2,H3,H4,H5,H6,P,CENTER,TD,TH,UL,DL,DIV {
	font-family: Geneva, Arial, Helvetica, sans-serif;
}
HR,BODY {
  margin-left: 4px;
  margin-right: 4px;
}
H1,H2,H3 {
	margin-bottom: 0px;
	padding-bottom: 0px;
	text-align: left;
	color: #000099;
}
/*BIG > STRONG { color: #000099; }*/
STRONG { color: #000099; }
P > STRONG { color: #000; }
H1 { margin-left: 4px; }
H2 { margin-left: 16px; }
H3 { margin-left: 32px; }
H4,H5,H6,P,CENTER,CL,DIV,TD,TH,DL,DT,DIV,UL {
	margin-left: 48px;
}
H1,H2,H3,H4,H5,H6,P,CENTER,CL,DIV,TD,TH,DL,DT,DIV,UL {
	margin-right: 24px;
}

DIV[name="index"] {
  margin: 0px;
  border: 1px solid #ccc;
  background-color: #f5f5f5;
}
DIV[name="index"] > p:first-child {
  margin: 0px;
  padding: 0px;
}
HR[name="index"], DIV[name="index"] + p {  display:none; }

UL UL { margin-left: 0px; }
LI > P  {
  margin-left: 0px;
}
P {
  margin-top: .25em;
  margin-bottom: .75em;
}

UL { list-style: square; }
UL UL { list-style: circle; }
UL UL UL { list-style: disc; }

PRE {
	border: 1px dashed #00f;
	background-color: #f5f5f5;
	margin-top: 4px;
	margin-bottom: 4px;
	margin-left: 48px;
	margin-right: 24px;
	padding-left: 0px;
	padding-right: 0px;
	padding-top: 4px;
	padding-bottom: 4px;
}
UL {
	/*background-color: #ff0;*/
	margin-top: 4px;
	margin-bottom: 4px;
}